(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: main.ml,v 1.10 2017/11/14 17:06:36 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module A = Uarray;;
module M = Umatrix;;
module Ar = Uaref;;
module Ok = Uoption_key;;
module Co = Ucommand_option;;
module U = Uunix;;

let _0_ = Nat._0_
and _1_ = Nat._1_
and _3_ = Nat._3_
and _5_ = Nat._5_
and _9_ = Nat._9_
;;

let piob = U.std_process_io_buffers;;

(* Shortcuts for the Pio library. *)

let pf = Pio.foreman_make_printing_facilities ();;

let wrap tw = Pio.worker_test_wrap_task piob tw;;

let make_input make_wrapper fun_name =
  let { Pio.tname = tname; inj = inj; _; } = make_wrapper pf fun_name in
  wrap tname inj
;;

(* Shortcuts for the Refinement library. *)

module W = Wrapper;;

let make_init_input = make_input W.make_init_wrapper "init"
and make_grad_input = make_input W.make_grad_wrapper "grad"
and make_optim_input = make_input W.make_optim_wrapper "optim"
and make_final_input = make_input W.make_final_wrapper "final"
;;

(* Main program for the generation of input data for the identity worker. *)

(* Command-line options. *)

let init_dim = _0_;;

let (get_dim, set_dim) =
  let { Ar.get; set; } = Ar.make_aref_unique "dimension" init_dim in
  let set_dim i = set (Nat.of_int i) in
  (get, set_dim)
;;

let dim_rco =
  {
   Co.key = Ok.long_option_key_of_string "-dimension";
   spec = Arg.Int set_dim;
   doc = Printf.sprintf "<nat>  Set dimension to <nat>.";
  }
;;

let { Ar.get_flag = get_use_mask; set_flag = set_use_mask; _; } =
  Ar.make_flag ()
;;

let use_mask_rco =
  {
   Co.key = Ok.long_option_key_of_string "use mask";
   spec = Arg.Unit set_use_mask;
   doc = Printf.sprintf "\n\tSet mask usage on. Default is off.";
  }
;;

let add_options () =
  Co.application_add_all_rco [ dim_rco; use_mask_rco; ]
;;

add_options ();;

let sys_argv_0 = A.get Sys.argv _0_;;

Co.set_usage_msg
  (Format.sprintf
     "\n%s -- Generate identity worker input\n\n\
      Usage: %s [<option>...] -dimension <nat>\n\n\
      Options are:@."
     sys_argv_0 sys_argv_0)
;;

(* Main. *)

let main () =
  Co.parse ();

  let dim = get_dim () in

  let array_make x = A.make dim x in
  let len = if not (get_use_mask ()) then _9_ else _5_ in
  let matrix_make x = M.make dim len x in
  let matrix_init f = M.init dim len f in

  let constant_matrix = matrix_make in
  let two_zone_matrix j1 x0 x1 =
    matrix_init
      (fun _i j -> if j < j1 then x0 else x1) in

  make_init_input dim;

  let pm = constant_matrix 127. in
  make_grad_input pm;

  let ppm = two_zone_matrix _3_ _0_ _1_
  and maa = array_make [| 0.; 0.; |] in
  make_optim_input (ppm, maa);

  let pm = two_zone_matrix _3_ 0. 127. in
  make_final_input pm;
;;

main ();;
