(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix_graph.ml,v 1.19 2017/11/07 20:57:40 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module P = Upervasives;;
module A = Uarray;;
module M = Umatrix;;

let _0_ = Nat._0_
and _1_ = Nat._1_
;;

let array_sub_for_all p a ofs len =
  let rec loop j i =
    j = len || p (A.get a i) && loop (Nat.succ j) (Nat.succ i) in
  loop _0_ ofs
;;

(* Shortcuts for the Mapping library. *)

module Iidx = Uindex_index;;
module Mset = Umatrix_nonempty_set;;

module Set = Mset;;

(* Matrix graph operations. *)

let qual = P.qualify "Ugraph";;
let invalid_arg = P.invalid_arg_qual qual;;
let failwith = P.failwith_qual qual;;

type index = Set.index
and uindex = Set.uindex
;;

type domain = Set.t;;

type sub_domain = Set.nonempty_subset;;

type 'a images =
  | Constant of 'a
  | Images of 'a array
;;

type 'a t =
  {
   domain : domain;
   images : 'a images;
  }
;;

let same = ( == );;

let same_domain mg1 mg2 = Set.same mg1.domain mg2.domain;;

let equal_domain mg1 mg2 = Set.equal mg1.domain mg2.domain;;

let unique_domain mg1 mg2 =
  if not (equal_domain mg1 mg2) then
    failwith "unique_domain" "domains differ" else
    { mg2 with domain = mg1.domain; }
;;

let equal_images equal_image { images = img1; _; } { images = img2; _; } =
  match (img1, img2) with
  | (Constant a1, Constant a2) -> equal_image a1 a2
  | (Constant a, Images ys) | (Images ys, Constant a) ->
      A.for_all (equal_image a) ys
  | (Images ys1, Images ys2) -> A.for_all2 equal_image ys1 ys2
;;

(* Construction functions. *)

let detect_constant_images mg =
  match mg.images with
  | Constant _ -> mg
  | Images ys ->
      let len = A.length ys in
      if len = _0_ then assert false else
      (* Now, we have len > 0.
         Therefore, the following A.unsafe_get and Nat.pred will not fail. *)
      let y0 = A.unsafe_get ys _0_ in
      let p y = y = y0 in
      let b = array_sub_for_all p ys _1_ (Nat.pred len) in
      if b then { mg with images = Constant y0; } else mg
;;

let make m =
  let sz = M.size m in
  let dom = Set.make sz in
  let ys = M.to_array m in
  detect_constant_images { domain = dom; images = Images ys; }
;;

let constant dom y = { domain = dom; images = Constant y; };;

let unsafe_from_array dom ys =
  detect_constant_images { domain = dom; images = Images ys; }
;;

let from_array dom ys =
  if Set.cardinal dom <> A.length ys then
    invalid_arg "from_array" "arguments have different sizes" else
  unsafe_from_array dom ys
;;

let from_fun dom f =
  let card = Set.cardinal dom in
  let (idx0, uidx0) = Set.choice dom in
  let y0 = f idx0 uidx0 in
  let ys = A.make card y0 in
  let ff idx uidx = A.set ys (Iidx.to_nat idx) (f idx uidx) in
  Set.sub_iter ff dom Iidx._1_ (Nat.pred card);
  unsafe_from_array dom ys
;;

(* Access functions. *)

let aux_constant fun_name mg f =
  match mg.images with
  | Constant y -> f mg y
  | Images _ -> invalid_arg fun_name "non constant graph"
;;

let get_constant mg = aux_constant "get_constant" mg (fun _ y -> y);;

let change_constant mg y =
  aux_constant "change_constant"
    mg (fun mg _ -> { mg with images = Constant y; })
;;

let image mg =
  match mg.images with
  | Constant y -> (fun _idx -> y)
  | Images ys -> (fun idx -> A.get ys (Iidx.to_nat idx))
;;

let to_array = function
  | { domain = dom; images = Constant y; } -> A.make (Set.cardinal dom) y
  | { images = Images a; _; } -> a
;;

let restriction
    ({ domain = dom; images = images; }  as mg)
    { Set.support = sup; indices = indices; nonempty_set = sub_dom; } =
  if sup != dom then
    invalid_arg "restriction" "sub-domain is not a subset of domain" else
  match images with
  | Constant y -> constant sub_dom y
  | Images _ -> from_array sub_dom (A.map (image mg) indices)
;;

(* Iterators. *)

let map_images f mg =
  match mg.images with
  | Constant y -> { mg with images = Constant (f y); }
  | Images ys -> { mg with images = Images (A.map f ys); }
;;

let fold_images f accu mg =
  match mg.images with
  | Constant y -> Set.fold_left (fun accu _ _ -> f accu y) accu mg.domain
  | Images ys -> A.fold_left f accu ys
;;
