(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix_numbering.ml,v 1.18 2017/11/07 20:57:40 fclement Exp $*)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module P = Upervasives;;
module A = Uarray;;
module M = Umatrix;;

let _0_ = Nat._0_;;

(* Shortcuts for the Mapping library. *)

module Num = Unumbering;;
module Midx = Umatrix_index;;
module Mmask = Umatrix_mask;;

module Mask = Mmask;;

(* Matrix numbering bijection operations. *)

let qual = P.qualify "Umatrix_numbering";;
let invalid_arg = P.invalid_arg_qual qual;;

type size = Mask.size;;

type key = Midx.element;;
type value = Midx.index;;

type t =
  {
   size : size;
   numbering : key Num.t;
  }
;;

(* Construction functions. *)

let valid_key (ni, nj) (i, j) = i < ni && j < nj;;

let check_key_in_bounds fname sz (i, j as k) =
  if not (valid_key sz k) then
      invalid_arg fname
        (Printf.sprintf "key (%d, %d) is out of bounds"
           (Nat.to_int i)
           (Nat.to_int j))
;;

let is_key_masked sm (i, j) = M.get sm i j = P.Forget;;

let check_key_non_masked fname sm (i, j as k) =
  if is_key_masked sm k then
    invalid_arg fname
      (Printf.sprintf "key (%d, %d) is masked"
        (Nat.to_int i)
        (Nat.to_int j))
;;

let check_key fname sz sm k =
  check_key_in_bounds fname sz k;
  check_key_non_masked fname sm k
;;

let valid_value len v = Midx.to_nat v < len;;

let check_value fname len v =
  if not (valid_value len v) then
    invalid_arg fname
      (Printf.sprintf "value %d out of bounds" (Midx.to_int v))
;;

let set_h h_dir_m h_inv_a (i, j as k) v =
  M.set h_dir_m i j v;
  A.set h_inv_a (Midx.to_nat v) k;
;;

let make_masked sm =
  let (ni, nj as sz) = M.size sm in
  let len = Mask.selection_length sm in
  let count_keep = ref Midx._0_ in
  let h_dir_m = M.make ni nj Midx._0_
  and h_inv_a = A.make len (_0_, _0_) in
  M.iter_ij
    (fun i j -> function
      | P.Forget ->
          M.set h_dir_m i j !count_keep
      | P.Keep ->
          set_h h_dir_m h_inv_a (i, j) !count_keep;
          Midx.incr count_keep)
    sm;
  let h_dir (i, j as k) =
    check_key "h_dir" sz sm k;
    M.get h_dir_m i j in
  let h_inv v =
    check_value "h_inv" len v;
    Midx.array_get h_inv_a v in
  { size = sz; numbering = Num.make len h_dir h_inv; }
;;

let of_mask mask =
  let sm = Mask.selection_matrix mask in
  make_masked sm
;;

let make sz = of_mask { Mask.size = sz; body = Mask.Empty; };;

(* Access functions. *)

let length mn = Num.length mn.numbering;;

let size mn = mn.size;;

let mem mn = valid_key mn.size;;

let value_of_key mn = Num.value_of_key mn.numbering;;

let key_of_value mn = Num.key_of_value mn.numbering;;

(* Iterators. *)

let iter f mn = Num.iter f mn.numbering;;

let fold_left f accu mn = Num.fold_left f accu mn.numbering;;
