(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: display_quantity_components.ml,v 1.35 2017/11/15 21:42:54 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module A = Uarray;;
module M = Umatrix;;
module Idx = Uindex;;
module Prat = Upositive_rational;;

let _0_ = Nat._0_;;

(* Shortcuts for Graphics/Grimage libraries. *)

let graphics_draw_rect (x, y) (w, h) =
  Graphics.draw_rect
    (Nat.to_int x) (Nat.to_int y) (Nat.to_int w) (Nat.to_int h)
;;

let graphics_moveto (x, y) = Graphics.moveto (Nat.to_int x) (Nat.to_int y);;

let pixmap_draw pxm (x, y) =
  Pixmap.draw_pixmap pxm (Nat.to_int x) (Nat.to_int y)
;;

(* Shortcuts for the Basement library. *)

module A_ng = Uarray_ng;;
module Nplane = Uaffine_plane.Natural;;

(* Shortcuts for the Mapping library. *)

module Mset = Umatrix_nonempty_set;;
module Mg = Umatrix_graph;;

module Set = Mset;;
module Map = Mg;;

(* Shortcuts for the Refinement library. *)

module Config = Configuration;;

open Refinement_types;;
open Display_types;;
open Display_types_print;;
open Display_window;;

(* Display quantity component operations. *)

(* Linear interpolation shortcut. *)

(* Type is nat -> nat -> float -> float -> int -> float. *)
let linear_interp i j c d =
  let inv_j_minus_i = 1. /. (Nat.to_float j -. Nat.to_float i) in
  let d_minus_c = d -. c in
  (fun x ->
    let ksi = (Nat.to_float x -. Nat.to_float i) *. inv_j_minus_i in
    c +. ksi *. d_minus_c)
;;

(* Graphics/Grimage shortcuts. *)

let get_blank () =
  let { window_background; _; } = get_graphics_elements () in
  window_background
;;

let display_rectangle w_sz pt = graphics_draw_rect pt w_sz;;

let display_string s pt =
  graphics_moveto pt;
  Graphics.draw_string s;
;;

let pixmap_make (w, h) c = M.to_array_array (M.make h w c);;
let pixmap_init (w, h) f = M.to_array_array (M.init h w f);;

let pixmap_draw sz pxm pt =
  pixmap_draw pxm pt;
  display_rectangle sz pt;
;;

(* Lower-left corner basics (see notes). *)

let shift_margin m pt = Nplane.add pt (m, m);;

let q_sw_llc q gm comps_dir (q_sw_w, q_sw_h) =
  let to_q_sw_llc =
    match (gm, q, comps_dir) with
    | (No_graphics, _, _) -> assert false
    | (Parameter_graphics, Param, _) -> Nplane.null
    | (Parameter_graphics, Grad, _) -> assert false
    | (Full_graphics, Param, Horizontal_comps_dir) -> (_0_, q_sw_h)
    | (Full_graphics, Param, Vertical_comps_dir) -> Nplane.null
    | (Full_graphics, Grad, Horizontal_comps_dir) -> Nplane.null
    | (Full_graphics, Grad, Vertical_comps_dir) -> (q_sw_w, _0_) in
  Nplane.add Nplane.origin to_q_sw_llc
;;

let pxms_sw_llc q gm comps_dir q_sw_sz cm_sw_t =
  let q_sw_llc = q_sw_llc q gm comps_dir q_sw_sz in
  let to_pxms_sw_llc =
    match comps_dir with
    | Horizontal_comps_dir -> Nplane.null
    | Vertical_comps_dir -> (_0_, cm_sw_t) in
  Nplane.add q_sw_llc to_pxms_sw_llc
;;

let pxm_sw_llc compl i q gm comps_dir q_sw_sz cm_sw_t (pxm_sw_w, pxm_sw_h) =
  let pxms_sw_llc = pxms_sw_llc q gm comps_dir q_sw_sz cm_sw_t in
  let to_pxm_sw_llc =
    match comps_dir with
    | Horizontal_comps_dir -> (Nat.mul i pxm_sw_w, _0_)
    | Vertical_comps_dir ->(_0_, Nat.mul (Nat.pred (compl i)) pxm_sw_h) in
  Nplane.add pxms_sw_llc to_pxm_sw_llc
;;

let pxm_llc compl i q gm comps_dir m =
  let { sub_window_sizes =
        { colormap_sub_window_tangential = cm_sw_t;
          pixmap_sub_window_size = pxm_sw_sz;
          quantity_sub_window_size = q_sw_sz; _; }; _; } =
    get_graphics_elements () in
  let pxm_sw_llc =
    pxm_sw_llc compl i q gm comps_dir q_sw_sz cm_sw_t pxm_sw_sz in
  shift_margin m pxm_sw_llc
;;

let tit_sw_llc q gm comps_dir ((_, q_sw_h) as q_sw_sz) tit_sw_h =
  let q_sw_llc = q_sw_llc q gm comps_dir q_sw_sz in
  let to_tit_sw_llc = (_0_, Nat.sub q_sw_h tit_sw_h) in
  Nplane.add q_sw_llc to_tit_sw_llc
;;

let tit_llc q gm comps_dir m =
  let { sub_window_sizes =
        { title_sub_window_height = tit_sw_h;
          quantity_sub_window_size = q_sw_sz; _; }; _; } =
    get_graphics_elements () in
  let tit_sw_llc = tit_sw_llc q gm comps_dir q_sw_sz tit_sw_h in
  shift_margin m tit_sw_llc
;;

let cm_sw_llc q gm comps_dir ((q_sw_w, _) as q_sw_sz) cm_sw_t =
  let q_sw_llc = q_sw_llc q gm comps_dir q_sw_sz in
  let to_cm_sw_llc =
    match comps_dir with
    | Horizontal_comps_dir -> (Nat.sub q_sw_w cm_sw_t, _0_)
    | Vertical_comps_dir -> Nplane.null in
  Nplane.add q_sw_llc to_cm_sw_llc
;;

let cb_sw_llc q gm comps_dir q_sw_sz cm_sw_t (leg_sw_t, _) =
  let cm_sw_llc = cm_sw_llc q gm comps_dir q_sw_sz cm_sw_t in
  let to_cb_sw_llc =
    match comps_dir with
    | Horizontal_comps_dir -> Nplane.null
    | Vertical_comps_dir -> (_0_, leg_sw_t) in
  Nplane.add cm_sw_llc to_cb_sw_llc
;;

let cb_llc q gm comps_dir m =
  let { sub_window_sizes =
        { legend_sub_window_polar_size = leg_sw_psz;
          colormap_sub_window_tangential = cm_sw_t;
          quantity_sub_window_size = q_sw_sz; _; }; _; } =
    get_graphics_elements () in
  let cb_sw_llc = cb_sw_llc q gm comps_dir q_sw_sz cm_sw_t leg_sw_psz in
  shift_margin m cb_sw_llc
;;

let cm_min_sw_llc q gm comps_dir q_sw_sz cm_sw_t cb_sw_t =
  let cm_sw_llc = cm_sw_llc q gm comps_dir q_sw_sz cm_sw_t in
  let to_cm_min_sw_llc =
    match comps_dir with
    | Horizontal_comps_dir -> (cb_sw_t, _0_)
    | Vertical_comps_dir -> Nplane.null in
  Nplane.add cm_sw_llc to_cm_min_sw_llc
;;

let cm_min_llc q gm comps_dir m =
  let { sub_window_sizes =
        { colorbar_sub_window_tangential = cb_sw_t;
          colormap_sub_window_tangential = cm_sw_t;
          quantity_sub_window_size = q_sw_sz; _; }; _; } =
    get_graphics_elements () in
  let cm_min_sw_llc = cm_min_sw_llc q gm comps_dir q_sw_sz cm_sw_t cb_sw_t in
  shift_margin m cm_min_sw_llc
;;

let cm_max_sw_llc
    q gm comps_dir q_sw_sz cm_sw_t cb_sw_t
    (pxm_sw_w, pxm_sw_h) (_, leg_sw_n) =
  let cm_min_sw_llc = cm_min_sw_llc q gm comps_dir q_sw_sz cm_sw_t cb_sw_t in
  let to_cm_max_sw_llc =
    match comps_dir with
    | Horizontal_comps_dir -> (_0_, Nat.sub pxm_sw_h leg_sw_n)
    | Vertical_comps_dir -> (Nat.sub pxm_sw_w leg_sw_n, _0_) in
  Nplane.add cm_min_sw_llc to_cm_max_sw_llc
;;

let cm_max_llc q gm comps_dir m =
  let { sub_window_sizes =
        { legend_sub_window_polar_size = leg_sw_psz;
          colorbar_sub_window_tangential = cb_sw_t;
          colormap_sub_window_tangential = cm_sw_t;
          pixmap_sub_window_size = pxm_sw_sz;
          quantity_sub_window_size = q_sw_sz; _; }; _; } =
    get_graphics_elements () in
  let cm_max_sw_llc =
    cm_max_sw_llc
      q gm comps_dir q_sw_sz cm_sw_t cb_sw_t pxm_sw_sz leg_sw_psz in
  shift_margin m cm_max_sw_llc
;;

(* Initialization. *)

let init_display_quantity_title q gm comps_dir m =
  let tit_llc = tit_llc q gm comps_dir m in
  let title =
    match q with
    | Param -> "Parameter"
    | Grad -> "Gradient" in
  display_string title tit_llc;
;;

let upward_fill (_w, h) min max map i _j =
  let xi = linear_interp _0_ h max min i in
  map xi
;;

let rightward_fill (w, _h) min max map _i j =
  let xj = linear_interp _0_ w min max j in
  map xj
;;

let init_colorbar_pixmap comps_dir cb_sz
    { Colormapping.color_scale = cs;
      from_float = min;
      to_float = max;
      mapping = map; _; } =
  let fill =
    match comps_dir with
    | Horizontal_comps_dir -> upward_fill
    | Vertical_comps_dir -> rightward_fill in
  pixmap_init cb_sz (fill cb_sz min max (map cs min max))
;;

let init_display_quantity_colorbar q gm comps_dir m cm =
  let cb_llc = cb_llc q gm comps_dir m in
  let { sub_window_sizes = { colorbar_size = cb_sz; _; }; _; } =
    get_graphics_elements () in
  let cb_pxm = init_colorbar_pixmap comps_dir cb_sz cm in
  pixmap_draw cb_sz cb_pxm cb_llc;
;;

let string_of_float = Printf.sprintf "%-8.2g";;

let init_display_quantity_colormap_legend q gm comps_dir m
    { Colormapping.from_float = min; to_float = max; _; } =
  let cm_min_llc = cm_min_llc q gm comps_dir m
  and cm_max_llc = cm_max_llc q gm comps_dir m in
  display_string (string_of_float min) cm_min_llc;
  display_string (string_of_float max) cm_max_llc;
;;

let init_display_quantity_colormap q gm comps_dir m =
  let cm =
    match q with
    | Param -> Config.get_graphics_parameter_color_mapping ()
    | Grad -> Config.get_graphics_gradient_color_mapping () in
  init_display_quantity_colorbar q gm comps_dir m cm;
  init_display_quantity_colormap_legend q gm comps_dir m cm;
;;

let init_display_quantity q gm comps_dir m =
  init_display_quantity_title q gm comps_dir m;
  init_display_quantity_colormap q gm comps_dir m;
;;

let init_display gm comps_dir m =
  match gm with
  | No_graphics -> assert false
  | Parameter_graphics -> init_display_quantity Param gm comps_dir m;
  | Full_graphics ->
      init_display_quantity Param gm comps_dir m;
      init_display_quantity Grad gm comps_dir m;
;;

(* Print pixmap basics. *)

let component_name i = Printf.sprintf "comp #%d" (Nat.to_int i);;

(* Alter pixmap basics. *)

let set_pxm_pxl pxm c (i, j) = A_ng.array_array_set pxm i j c;;

(* Function scale is the multiplication by a rational number scaling factor.
 It transforms coordinates in the computed image into coordinates in the
 displayed image. Therefore, the rational coefficient is greater than 1 when
 the displayed image is larger than the computed image, and of course it is
 lesser than 1 otherwise. *)

let set_image_pixel set_pixel =
  let { scaling_factor = factor; _; } = get_graphics_elements () in
  let scale = Prat.scale factor in
  (fun (im_i, im_j) ->
     let imin = scale im_i
     and imax = Nat.pred (scale (Nat.succ im_i)) in
     let jmin = scale im_j
     and jmax = Nat.pred (scale (Nat.succ im_j)) in
     M.nat_for_loop2 (imin, jmin) (imax, jmax) set_pixel)
;;

let set_zone_pixel set_pixel stc imgs dom idx _uidx =
  let x =
    match imgs with
    | Map.Images ims -> A.get ims (Idx.to_nat idx)
    | Map.Constant x -> x in
  let c = stc x in
  let set_pixel = set_pixel c in
  let im_cell = Set.element dom idx in
  set_image_pixel set_pixel im_cell;
;;

let set_zone pxm stc _zn { Map.domain = dom; images = imgs; } =
  let set_pixel = set_pxm_pxl pxm in
  let set_zone_pixel = set_zone_pixel set_pixel stc imgs dom in
  Set.iter set_zone_pixel dom;
;;

module Make =
  functor (Scal : Scalar_spec.SCALAR) ->
  functor (Comps : Components_spec.COMPONENTS) ->
struct
  module H = History.Make (Scal) (Comps);;
  module Bb = H.Bb;;
  module St = H.St;;

  module Dqt = Display_quantity_types.Make (Scal) (Comps);;

  let compl =
    let n = Comps.get_dim () in
    (fun i -> Nat.sub n i)
;;

  let fk_comps = Comps.make;;

  let mk_pq _fake comps _gm = comps;;

  let mk_gq fake comps = function
    | No_graphics -> assert false
    | Parameter_graphics -> fake
    | Full_graphics -> comps
;;

  let mk_xq = function
    | Param -> mk_pq
    | Grad -> mk_gq
;;

  let mk_comps q fk = (mk_xq q) (fk_comps fk);;

  let get_comps comps = Comps.map_rails comps (Comps.make ());;

  (* For late enough evaluation of its components, it is easier to only
   consider functional comps. If necessary, thunks are added, i.e. for
   pixmaps, and lower_left_corners. *)

  (* Pixmaps. *)

  let fk_get_pxm = (fun () -> [||]);;

  let mk_get_pxm d_sz = (fun () -> pixmap_make d_sz (get_blank ()));;
  let mk_get_pxm_comps d_sz = Comps.make (mk_get_pxm d_sz);;

  let mk_get_pxms q gm d_sz =
    let fk = fk_get_pxm
    and comps = mk_get_pxm_comps d_sz in
    mk_comps q fk comps gm
;;

  let make_pixmaps q gm d_sz = get_comps (mk_get_pxms q gm d_sz);;

  (* Lower_left_corners. *)

  let fk_get_ll = (fun () -> Nplane.origin);;

  let mk_get_ll i q gm comps_dir m =
    (fun () -> pxm_llc compl i q gm comps_dir m)
;;
  let mk_get_ll_comps q gm comps_dir m =
    Comps.init (fun i -> mk_get_ll i q gm comps_dir m)
;;

  let mk_get_lls q gm comps_dir m =
    let fk = fk_get_ll
    and comps = mk_get_ll_comps q gm comps_dir m in
    mk_comps q fk comps gm
;;

  let make_lower_left_corners q gm comps_dir m =
    get_comps (mk_get_lls q gm comps_dir m)
;;

  (* Print_pixmaps. *)

  let fk_prt = (fun _ppf -> ());;

  let fake_print_pixmaps = fk_comps fk_prt;;

  let mk_prt q s pxm ppf =
    Format.fprintf ppf "@[<hv 2>%a %s pxm =@ %a@]"
      print_quantity q s print_pixmap pxm
;;
  let mk_prt_comps q = Comps.map2 (mk_prt q) (Comps.init component_name);;

  let make_print_pixmaps q pxms gm =
    let fk = fk_prt
    and comps = mk_prt_comps q pxms in
    mk_comps q fk comps gm
;;

  (* Alter_pixmaps. *)

  let fk_alt = (fun _iter _stc _pp -> ());;

  (* We need to contraint the type of fake_alter_pixmaps since the version
     4.05 of the OCaml compiler. *)
  let (fake_alter_pixmaps : Dqt.scalar_alter_pixmaps) = fk_comps fk_alt;;

  let mk_alt pxm iter stc pp = iter (set_zone pxm stc) pp;;
  let mk_alt_comps pxms = Comps.map mk_alt pxms;;

  let make_alter_pixmaps q pxms gm =
    let fk = fk_alt
    and comps = mk_alt_comps pxms in
    mk_comps q fk comps gm
;;

  (* Draw_pixmaps. *)

  let fk_drw = (fun () -> ());;

  let fake_draw_pixmaps = fk_comps fk_drw;;

  let mk_drw d_sz pxm ll = (fun () -> pixmap_draw d_sz pxm ll);;
  let mk_drw_comps d_sz = Comps.map2 (mk_drw d_sz);;

  let make_draw_pixmaps q d_sz pxms lls gm =
    let fk = fk_drw
    and comps = mk_drw_comps d_sz pxms lls in
    mk_comps q fk comps gm
;;

  (* Application. *)

  let apply_comps apply rc = Comps.iter apply rc;;

  let print_apply print ppf = apply_comps (fun f -> print f ppf)
  and draw_apply = apply_comps (fun f -> f ())
;;

  let alter_apply iter stcs pps comps =
    let comps = Comps.map (fun f -> f iter) comps in
    Comps.iter_rails2 comps stcs pps;
;;
end
;;
