(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: state.ml,v 1.47 2017/11/14 17:06:35 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module L = Ulist;;
module A = Uarray;;
module Dgraph = Udynamic_graph;;

let _0_ = Nat._0_;;

(* Shortcuts for the Basement library. *)

module Alist = Uassoc_list;;

(* Shortcuts for the Mapping library. *)

module Mset = Umatrix_nonempty_set;;
module Mg = Umatrix_graph;;

module Set = Mset;;
module Map = Mg;;

(* Shortcuts for the Refinement library. *)

module Config = Configuration;;
module Ostp = Optimization_status_types_print;;

open Refinement_types;;
open Refinement_types_print;;
open Stopping_criterion_types_print;;

(* State operations. *)

module Make =
  functor (Scal : Scalar_spec.SCALAR) ->
  functor (Comps : Components_spec.COMPONENTS) ->
struct
  module Bb = Black_box.Make (Scal) (Comps);;
  module Stt = Bb.Stt;;

  type scalar = Scal.t;;
  type 'a components = 'a Comps.t;;
  type t = Bb.state;;
  type selected_refinement = Bb.selected_refinement;;
  type constants = Constants_types.t;;

  type scalar_forced_zonation = scalar forced_zonation;;
  type forced_zonations = scalar_forced_zonation components;;

  let same
      { Stt.zonations = zs1; cost = j1; _; }
      { Stt.zonations = zs2; cost = j2; _; } =
    j1 == j2 && Comps.same Zonation.same zs1 zs2
  ;;

  let equal_coarse_parameters pps { Stt.zonations = zs; _; } =
    Comps.for_all2 Zonation.equal_coarse_parameter pps zs
  ;;

  let all_equal_coarse_parameters ppsl srl =
    L.for_all2 equal_coarse_parameters ppsl srl
  ;;

  let printer_zonations = Comps.printer (printer_zonation Scal.printer);;

  let printer_poly_state pp_a ctx ppf = function
  | { Stt.stopping_criterion = a; zonations = z; cost = c; } ->
    if ctx = Og_precedence.Ck_argument || ctx = Og_precedence.Ck_record_field
      then Format.fprintf ppf "%, {@ " else
    Format.fprintf ppf "%,@[<hv 2>{@ "; Format.fprintf ppf "%,@[<hv 0>";
    Format.fprintf ppf "%,Stt.@,";
    Format.fprintf ppf "%,@[<hv 2>stopping_criterion =";
    pp_a Og_precedence.Ck_record_field ppf a; Format.fprintf ppf "%,;@]";
    Format.fprintf ppf "%,@ "; Format.fprintf ppf "%,@[<hv 2>zonations =";
    printer_zonations Og_precedence.Ck_record_field ppf z;
    Format.fprintf ppf "%,;@]"; Format.fprintf ppf "%,@ ";
    Format.fprintf ppf "%,@[<hv 2>cost =";
    printer_cost Og_precedence.Ck_record_field ppf c;
    Format.fprintf ppf "%,;@]"; Format.fprintf ppf "%,@]";
    if ctx = Og_precedence.Ck_argument || ctx = Og_precedence.Ck_record_field
      then Format.fprintf ppf "%,@;<1 -2>}" else
    Format.fprintf ppf "%,@;<1 -2>}@]"
  ;;

  let printer_state ctx ppf = function
    | s_p -> printer_poly_state printer_stopping_criterion ctx ppf s_p
  ;;

  let printer_selected_refinement =
    printer_poly_state Ostp.printer_optimization_status
  ;;

  let printer = printer_state;;
  let print = printer Og_precedence.Ck_toplevel;;

  let print_selected_refinement =
    printer_selected_refinement Og_precedence.Ck_toplevel
  ;;

  let printer_constants = Constants_types_print.printer_t;;
  let print_constants = Constants_types_print.print_t;;

  (* Only used for cold initialization. *)
  let init = Bb.init_state;;
  let init_constants = Bb.init_constants;;

  let vector_first_order_indicators vfoi_add_zones fzs =
    let fz0 = Comps.get fzs _0_ in
    let accu = Scal.vfoi_init fz0 in
    let fz = Comps.fold_left vfoi_add_zones accu fzs in
    Comps.make fz
  ;;

  let vector_first_order_indicators_overall_best =
    vector_first_order_indicators Scal.vfoi_add_zones_overall_best
  and vector_first_order_indicators_best_in_a_family =
    vector_first_order_indicators Scal.vfoi_add_zones_best_in_a_family
  ;;

  let dm_treatment f accu fzs =
    match Config.get_dimension_management () with
    | Best_component_only -> Comps.fold_left_i f accu fzs
    | Vector -> f accu _0_ (Comps.get fzs _0_)
  ;;

  let max_foi_cc accu { first_order_indicator = foi; _; } = max accu foi;;

  let max_foi_zone accu { Dgraph.image = ccl; _; } =
    L.fold_left max_foi_cc accu ccl
  ;;

  let max_foi_scalar accu _cn = A.fold_left max_foi_zone accu;;

  let max_first_order_indicator fzs =
    dm_treatment max_foi_scalar Scal._0_ fzs
  ;;

  let init_partition len zn (sst, ssf) fz i =
    (* init_partition is called with len >= 1.
       Therefore, the following Nat.pred will not fail. *)
    let imax = Nat.pred len in
    let i' = if i = imax then zn else i in
    let fsmi' = Dgraph.get_value (A.get fz i') in
    let mi' = Map.get_constant fsmi' in
    let constant_sm ss = Map.constant ss.Set.nonempty_set mi' in
    if i = imax then constant_sm sst else
    if i = zn then constant_sm ssf else
    fsmi'
  ;;

  let split_scalar zn ct fz =
    let len = Nat.succ (A.length fz) in
    (* We always have len >= 1. *)
    A.init len (init_partition len zn ct fz)
  ;;

  let no_split_scalar fz = A.map Dgraph.get_value fz;;

  let make_refinement cn zn ct =
    match Config.get_dimension_management () with
    | Best_component_only ->
        Comps.singleton cn (split_scalar zn ct) no_split_scalar
    | Vector ->
        Comps.init (fun _ -> split_scalar zn ct)
  ;;

  let split cn zn ct fzs =
    let refinements = make_refinement cn zn ct in
    Comps.map_rails refinements fzs
  ;;

  let select_cc
      fzs select cn zn accu
      { refinement = ct; first_order_indicator = foi; _; } =
    if select foi then (split cn zn ct fzs, foi) :: accu else accu
  ;;

  let select_zone fzs select cn accu zn fzone =
    L.fold_left (select_cc fzs select cn zn) accu fzone.Dgraph.image
  ;;

  let select_scalar fzs select accu cn fz =
    A.fold_left_i (select_zone fzs select cn) accu fz
  ;;

  let select_n_best_cc fzs select nb =
    let ppsl = dm_treatment (select_scalar fzs select) [] fzs in
    Alist.keys_with_highest_values nb ppsl
  ;;

  let final = Bb.final;;
end
;;
